@@ -1,10 +1,7 @@
-/* pam_ftp module */
-
 /*
- * $Id$
+ * pam_ftp module
  *
  * Written by Andrew Morgan <morgan@linux.kernel.org> 1996/3/11
- *
  */
 
 #define PLEASE_ENTER_PASSWORD "Password required for %s."
@@ -23,18 +20,10 @@
 #include <stdarg.h>
 #include <string.h>
 
-/*
- * here, we make a definition for the externally accessible function
- * in this file (this definition is required for static a module
- * but strongly encouraged generally) it is used to instruct the
- * modules include file to define the function prototypes.
- */
-
-#define PAM_SM_AUTH
-
 #include <security/pam_modules.h>
 #include <security/_pam_macros.h>
 #include <security/pam_ext.h>
+#include "pam_inline.h"
 
 /* argument parsing */
 
@@ -49,18 +38,18 @@ _pam_parse(pam_handle_t *pamh, int argc, const char **argv, const char **users)
 
     /* step through arguments */
     for (ctrl=0; argc-- > 0; ++argv) {
+	const char *str;
 
 	/* generic options */
 
 	if (!strcmp(*argv,"debug"))
 	    ctrl |= PAM_DEBUG_ARG;
-	else if (!strncmp(*argv,"users=",6)) {
-	    *users = 6 + *argv;
-	} else if (!strcmp(*argv,"ignore")) {
+	else if (!strcmp(*argv,"ignore"))
 	    ctrl |= PAM_IGNORE_EMAIL;
-	} else {
+	else if ((str = pam_str_skip_prefix(*argv, "users=")) != NULL)
+	    *users = str;
+	else
 	    pam_syslog(pamh, LOG_ERR, "unknown option: %s", *argv);
-	}
     }
 
     return ctrl;
@@ -122,7 +111,7 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
     const char *users = NULL;
 
     /*
-     * this module checks if the user name is ftp or annonymous. If
+     * this module checks if the user name is ftp or anonymous. If
      * this is the case, it can set the PAM_RUSER to the entered email
      * address and SUCCEEDS, otherwise it FAILS.
      */
@@ -130,8 +119,9 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
     ctrl = _pam_parse(pamh, argc, argv, &users);
 
     retval = pam_get_user(pamh, &user, NULL);
-    if (retval != PAM_SUCCESS || user == NULL) {
-	pam_syslog(pamh, LOG_ERR, "no user specified");
+    if (retval != PAM_SUCCESS) {
+	pam_syslog(pamh, LOG_NOTICE, "cannot determine user name: %s",
+		   pam_strerror(pamh, retval));
 	return PAM_USER_UNKNOWN;
     }
 
@@ -143,6 +133,8 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 	retval = pam_set_item(pamh, PAM_USER, (const void *)anon_user);
 	if (retval != PAM_SUCCESS || anon_user == NULL) {
 	    pam_syslog(pamh, LOG_ERR, "user resetting failed");
+	    free(anon_user);
+
 	    return PAM_USER_UNKNOWN;
 	}
 	free(anon_user);
@@ -165,7 +157,7 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 			       GUEST_LOGIN_PROMPT);
 
 	if (retval != PAM_SUCCESS) {
-	    _pam_overwrite (resp);
+	    pam_overwrite_string (resp);
 	    _pam_drop (resp);
 	    return ((retval == PAM_CONV_AGAIN)
 		    ? PAM_INCOMPLETE:PAM_AUTHINFO_UNAVAIL);
@@ -185,7 +177,7 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 		}
 	    }
 
-	    /* we are happy to grant annonymous access to the user */
+	    /* we are happy to grant anonymous access to the user */
 	    retval = PAM_SUCCESS;
 
 	} else {
@@ -204,7 +196,7 @@ pam_sm_authenticate (pam_handle_t *pamh, int flags UNUSED,
 	}
 
 	/* clean up */
-	_pam_overwrite(resp);
+	pam_overwrite_string(resp);
 	_pam_drop(resp);
 
 	/* success or failure */
